var fs = require('fs'),
    path = require('path'),
    tarima = require('tarima');

module.exports = function(grunt) {
  var _ = grunt.util._;

  grunt.registerMultiTask('tarima', 'Compile Tarima templates into JavaScript or plain output', function() {
    var options = this.options({
      callback: null,
      join: false,
      timeout: 260,
      includes: {},
      locals: {},
      dest: '',
      cwd: '',
      raw: []
    });

    _.each(options.raw, function(type) {
      tarima.add(type, function(params) {
        return params.source;
      });
    })

    _.each(options.includes, function(all, type) {
      _.each(all, function(file, index) {
        options.includes[type][index] = grunt.file.read(file);
      });
    });

    this.files.forEach(function (f) {
      var validFiles = removeInvalidFiles(f.src);

      if ('function' === typeof options.filter) {
        validFiles = options.filter(validFiles);
      }

      writeFile(options.dest || f.dest, concatOutput(validFiles, options), options);
    });
  });

  var removeInvalidFiles = function(files) {
    return files.filter(function(filepath) {
      if (!grunt.file.exists(filepath)) {
        grunt.verbose.warn('Source file "' + filepath + '" not found.');
        return false;
      } else {
        var exts = filepath.split('/').pop().split('.');

        if (exts.length < 2) {
          return false;
        }

        if (['jade', 'ract', 'less', 'hbs', 'us'].indexOf(exts[1]) > -1) {
          grunt.verbose.writeln('Source file "' + filepath + '" skipped.');
          return false;
        }

        return true;
      }
    });
  };

  var concatOutput = function(files, options) {
    var fixed = [];

    files.forEach(function(filepath) {
      var peak = filepath.split('.').pop(),
          params = _.omit(options, 'filter', 'callback', 'join');

      if (peak === filepath) {
        return;
      }

      if (filepath.indexOf(':') > -1) {
        fixed.push(compileTarima(grunt.file.read(filepath.split(':')[0]), params, filepath.split(':')[1]))
      } else {
        fixed.push(compileTarima(grunt.file.read(filepath), params, filepath));
      }
    });

    return fixed;
  };

  var compileTarima = function(code, options, filepath) {
    try {
      return tarima.parse(filepath, code, options);
    } catch (e) {
      grunt.fail.warn('Tarima failed to compile.');
    }
  };

  var warnOnEmptyFile = function (path) {
    grunt.log.warn('File "' + path + '" not written because is empty.');
  };

  var writeFile = function (path, files, options) {
    var filepath, partial, code, next;
    var views = {}, exts = [], plain;

    if (!files.length) {
      grunt.log.ok('Without changes.');
      return;
    }

    files.forEach(function(partial) {
      if ('function' === typeof options.process) {
        partial = options.process(partial);
      }

      next = partial.params.parts[0];
      plain = 'js' !== partial.params.ext;
      code = partial.compile(options.locals);

      if (!code || code.length < 1) {
        warnOnEmptyFile(path + partial.params.keypath + '.' + partial.params.ext);
      } else {
        next = next || partial.params.ext;

        if (options.join) {
          if (exts.indexOf(next) === -1) {
            exts.push(next);
          }

          views[partial.params.keypath] = code;
        } else {
          path = path.substr(-1) === '/' ? path : path + '/';

          if (options.cwd) {
            filepath = partial.params.filepath.substr(options.cwd.length).replace(/^\//, '');
            filepath = filepath ? filepath + '/' + partial.params.name : partial.params.name;
            filepath = path + filepath;
          } else {
            filepath = path + partial.params.name;
          }

          if (plain) {
            filepath += '.' + partial.params.ext;
            grunt.file.write(filepath, code);
          } else {
            filepath += '.' + partial.params.ext;
            grunt.file.write(filepath, getTpl(code, getDeps(next), options.callback));
          }

          grunt.log.ok('Source file "' + filepath + '" created.');
        }
      }
    });

    if (options.join) {
      if (plain) {
        code = _.values(views).join('\n');
      } else {
        code = getTpl(views, _.map(_.unique(exts), function(value) { return getDeps(value); }).join(''), options.callback);
      }

      grunt.file.write(path, code);
      grunt.log.ok('File ' + path + ' created.');
    }
  };

  var fnWrapper = function(val, main) {
    var code = [],
        wrapped = false;

    if (main) {
      if (/(^|\n)[ \t]*(module.)?exports\s+=/.test(val)) {
        return val;
      }

      code.push('module.exports = ');
      wrapped = true;
    }

    if (!/^\s*(function[\s(]|Handlebars\.template)/.test(val)) {
      code.push('function(){' + val + '}');
    } else {
      code.push(val);
    }

    if (wrapped) {
      code.push(';');
    }

    return code.join('');
  };

  var fixDep = function(type) {
    var map = {
      hbs: 'require("handlebars/runtime")["default"]',
      jade: 'require("jade/runtime")'
    };

    return map[type] || ('require("' + type + '")');
  };

  var getDeps = function (type) {
    var prefix = '',
        context = "(typeof window !== 'undefined' ? window : global)";

    var html_escape = function(s) {
          return s !== null ? String(s).replace(/[&<>"']/g, function(m) {
            return ({ '&': '&amp;', '<': '&lt;', '>': '&gt;', '"': '&quot;', "'": '&#39;' })[m];
          }) : '';
        };

    switch (type) {
      case 'less':
      case 'jade':
        prefix = "var " + type + " = " + context + "." + type + " || " + fixDep(type) + ";";
      break;
      case 'hbs':
        prefix = "var Handlebars = " + context + ".Handlebars || " + fixDep(type) + ";";
      break;
      case 'us':
        prefix = "var _ = " + context + "._ || { escape: " + html_escape.toString() + " };";
      break;
    }

    return prefix;
  };

  var getTpl = function(tpl, deps, callback) {
    var output = [],
        views = [];

    if ('function' === typeof callback) {
      return callback(tpl, deps);
    }

    output.push(deps);

    if ('object' === typeof tpl) {
      output.push('module.exports = {');

      _.each(tpl, function(value, key) {
        views.push("\n'" + key + "': " + fnWrapper(value));
      });

      output.push(views.join(','));
      output.push('};');
    } else {
      output.push(fnWrapper(tpl, true));
    }

    return output.join('');
  };
};
